import pkg_resources
import numpy as np
import torch
import torch.nn.functional as F

_CACHE_LOCATION = '.cache'


def bitget(byteval, idx):
    return ((byteval & (1 << idx)) != 0)


def cache_location():
    return pkg_resources.resource_filename(__name__, _CACHE_LOCATION)


def compute_cm(gt, pred, N, with_cuda=False):
    '''Computes confusion matrix between prediction and gt
    Args:
      gt (Tensor) : gt.
      pred (Tensor) : predictions.
      N (int) : number of classes in matrix.
    Returns:
      Confusion matrix
      (Tensor of size (N, N)).
    '''
    # flatten gt and preds
    pred = pred.view([-1])
    gt = gt.view([-1])

    # check for valid labels
    mask_1 = gt >= 0
    mask_2 = gt < N
    mask = torch.logical_and(mask_1, mask_2)

    # compute scatter idx
    pred = pred[mask]
    gt = gt[mask]
    gt *= N
    idx = gt + pred

    cm = torch.zeros([N * N], dtype=torch.float32)
    if with_cuda:
        cm = cm.cuda()
    samples = torch.ones_like(gt, dtype=torch.float)

    # compute confusion matrix and reshape
    cm = cm.scatter_add_(0, idx, samples)
    cm = cm.view([N, N]).type(torch.int64)

    return cm


def compute_iu(cm):
    '''Compute IU from confusion matrix.
    Args:
      cm (ndarray) : square confusion matrix.
    Returns:
      IU vector (ndarray).
    '''
    N = cm.shape[0]
    IU = np.ones(N)

    for i in range(N):
        fn = np.sum(cm[:, i])
        fp = np.sum(cm[i, :])
        tp = cm[i, i]
        total = fn + fp - tp
        if total > 0:
            IU[i] = tp / total
    return IU


def forward_multi_scale(model, img):
    # sample multi scale images
    height = img.shape[-2]
    width = img.shape[-1]

    # scales
    scales = [0.4, 0.6, 0.8, 1, 1.2]

    # predict at multiple scales
    predictions = []
    for scale in scales:

        # compute scaled height and width
        scaled_height = int(scale * height)
        scaled_width = int(scale * width)

        # interpolate image to scale
        scaled_img = F.interpolate(img, (scaled_height, scaled_width),
                                   mode='bilinear')

        # forward pass through
        logits = model(scaled_img)

        # interpolate logits back to original image size
        prediction = F.softmax(logits, dim=1)
        prediction = F.interpolate(prediction, (height, width),
                                   mode='bilinear')
        predictions.append(prediction)

    prediction = torch.cat(predictions, dim=0)
    prediction = torch.mean(prediction, dim=0)

    # average across images
    prediction = torch.argmax(prediction, dim=0)
    prediction = torch.squeeze(prediction)

    return prediction


def forward_single_scale(model, img):
    # forward pass through
    logits = model(img)

    # interpolate logits back to original image size
    prediction = F.softmax(logits, dim=1)
    prediction = F.interpolate(prediction, (img.shape[-2], img.shape[-1]),
                               mode='bilinear')
    prediction = torch.argmax(prediction, dim=1)
    prediction = torch.squeeze(prediction)

    return prediction


class ColourMap(object):

    def __init__(self, N=256, normalised=False, dataset='voc'):
        dtype = 'float32' if normalised else 'uint8'

        if dataset == 'voc' or dataset == 'nyu':
            self.map = np.zeros((N, 3), dtype=dtype)
            for i in range(N):
                r = g = b = 0
                c = i
                for j in range(8):
                    r = r | (bitget(c, 0) << 7 - j)
                    g = g | (bitget(c, 1) << 7 - j)
                    b = b | (bitget(c, 2) << 7 - j)
                    c = c >> 3

                self.map[i] = np.array([r, g, b])
        else:
            self.map = np.array(
                [[128, 64, 128], [244, 35, 232], [70, 70, 70], [102, 102, 156],
                 [190, 153, 153], [153, 153, 153], [250, 170, 30], [
                     220, 220, 0
                 ], [107, 142, 35], [152, 251, 152], [70, 130, 180],
                 [220, 20, 60], [255, 0, 0], [0, 0, 142], [0, 0, 70],
                 [0, 60, 100], [0, 80, 100], [0, 0, 230], [119, 11, 32]],
                dtype=dtype)

        self.map = self.map / 255 if normalised else self.map

    def colourise(self, prediction):
        prediction = self.map[prediction]
        return prediction



# creates the mappings COCO to VOC classes
def coco2voc():
    voc2coco = [[0], [5], [2], [16], [9], [44], [6], [3, 8], [17], [62], [21],
                [67], [18], [19], [4], [1], [64], [20], [63], [7], [72]]
    coco2voc = [0] * 91
    for voc_idx in range(len(voc2coco)):
        for coco_idx in voc2coco[voc_idx]:
            coco2voc[coco_idx] = voc_idx

    return coco2voc


# creates the mappings of full Citiscapes classes to training classes
def cs2train():
    train2cs = [[7], [8], [11], [12], [13], [17], [19], [20], [21], [22], [23],
                [24], [25], [26], [27], [28], [31], [32], [33]]
    cs2train = [255] * 34
    for train_idx in range(len(train2cs)):
        for cs_idx in train2cs[train_idx]:
            cs2train[cs_idx] = train_idx

    return cs2train


# read list of files corresponding to train/val/test split
def read_filelist(file):
    file_list = []
    with open(file, 'r') as f:
        line = f.readline()
        while line:
            file_list.append(line.strip())
            line = f.readline()

    return file_list



import os
import numpy as np
import torch
import random
from PIL import Image
from scipy.io import loadmat
from torch.utils.data.dataset import Dataset

from .helpers import read_filelist
from ..helpers import ColourMap


class SBD(Dataset):
    '''Semantic Boundaries Segmentation dataset.'''
    COLOUR_MAP = ColourMap(dataset='voc')
    LABEL_OFFSET = 0
    NUM_CLASSES = 21

    def __init__(self,
                 root_dir,
                 image_set='train',
                 transform=None,
                 target_transform=None):
        '''
        Args:
            csv_file (string): Path to the csv file with annotations.
            root_dir (string): Directory with all the images.
            transform (callable, optional): Optional transform to be applied
                on a sample.
        '''

        self.root_dir = root_dir
        self.image_set = image_set
        if self.image_set == 'train':
            self.file_list = read_filelist(
                os.path.join(root_dir, 'benchmark_RELEASE', 'dataset',
                             'train.txt'))
        elif self.image_set == 'val':
            self.file_list = read_filelist(
                os.path.join(root_dir, 'benchmark_RELEASE', 'dataset',
                             'val.txt'))
        self.transform = transform
        self.target_transform = target_transform

        # dataset properties
        self.num_classes = SBD.NUM_CLASSES
        self.ignore_index = 255
        self.label_offset = SBD.LABEL_OFFSET
        self.cmap = SBD.COLOUR_MAP

    def __len__(self):
        return len(self.file_list)

    def __getitem__(self, idx):
        if torch.is_tensor(idx):
            idx = idx.tolist()

        # get filename
        filename = self.file_list[idx]

        # load image data
        img_name = os.path.join(self.root_dir, 'benchmark_RELEASE', 'dataset',
                                'img', filename + '.jpg')
        image = Image.open(img_name)
        image = image.convert('RGB')

        # load label data
        label_name = os.path.join(self.root_dir, 'benchmark_RELEASE',
                                  'dataset', 'cls', filename + '.mat')
        label = loadmat(label_name)
        label = label['GTcls']['Segmentation'][0][0]
        label = Image.fromarray(label)

        seed = np.random.randint(2147483647)
        random.seed(seed)
        if self.transform:
            image = self.transform(image)

        random.seed(seed)
        if self.target_transform:
            label = self.target_transform(label)

        # convert to label to tensor (without scaling to [0,1])
        label = np.asarray(label).astype(np.uint8)
        label = torch.from_numpy(label).type(torch.LongTensor)

        # create sample of data and label
        sample = {'name': filename, 'data': image, 'label': label}

        return sample



import torch.nn as nn
import torch.nn.functional as F

def conv3x3(in_planes, out_planes, stride=1, bias=False):
    return nn.Conv2d(in_planes, out_planes, kernel_size=3, stride=stride,
                     padding=1, bias=bias)

def conv1x1(in_planes, out_planes, stride=1, bias=False):
    return nn.Conv2d(in_planes, out_planes, kernel_size=1, stride=stride,
                     padding=0, bias=bias)

class BasicBlock(nn.Module):
    expansion = 1

    def __init__(self, inplanes, planes, stride=1, downsample=None):
        super(BasicBlock, self).__init__()
        self.conv1 = conv3x3(inplanes, planes, stride)
        self.bn1 = nn.BatchNorm2d(planes)
        self.relu = nn.ReLU(inplace=True)
        self.conv2 = conv3x3(planes, planes)
        self.bn2 = nn.BatchNorm2d(planes)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x):
        residual = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)

        if self.downsample is not None:
            residual = self.downsample(x)

        out += residual
        out = self.relu(out)

        return out


class Bottleneck(nn.Module):
    expansion = 4

    def __init__(self, inplanes, planes, stride=1, downsample=None):
        super(Bottleneck, self).__init__()
        self.conv1 = nn.Conv2d(inplanes, planes, kernel_size=1, bias=False)
        self.bn1 = nn.BatchNorm2d(planes)
        self.conv2 = nn.Conv2d(planes, planes, kernel_size=3, stride=stride,
                               padding=1, bias=False)
        self.bn2 = nn.BatchNorm2d(planes)
        self.conv3 = nn.Conv2d(planes, planes * 4, kernel_size=1, bias=False)
        self.bn3 = nn.BatchNorm2d(planes * 4)
        self.relu = nn.ReLU(inplace=True)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x):
        residual = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = self.relu(out)

        out = self.conv3(out)
        out = self.bn3(out)

        if self.downsample is not None:
            residual = self.downsample(x)

        out += residual
        out = self.relu(out)

        return out

class ChainedResidualPool(nn.Module):

    def __init__(self, in_planes, out_planes, n_stages):
        super(ChainedResidualPool, self).__init__()

        self.stride = 1
        self.n_stages = n_stages

        for i in range(n_stages):
            setattr(self, '{}_{}'.format(i + 1, 'outvar_dimred'),
                    conv3x3(in_planes if (i == 0) else out_planes, out_planes, stride=self.stride))
        self.maxpool = nn.MaxPool2d(kernel_size=5, stride=1, padding=2)

    def forward(self, x):
        top = x
        for i in range(self.n_stages):
            top = self.maxpool(top)
            top = getattr(self, '{}_{}'.format(i + 1, 'outvar_dimred'))(top)
            x = top + x
        return x

class ChainedResidualPoolLW(nn.Module):

    def __init__(self, in_planes, out_planes, n_stages):
        super(ChainedResidualPoolLW, self).__init__()

        self.stride = 1
        self.n_stages = n_stages

        for i in range(n_stages):
            setattr(self, '{}_{}'.format(i + 1, 'outvar_dimred'),
                    conv1x1(in_planes if (i == 0) else out_planes, out_planes, stride=self.stride))
        self.maxpool = nn.MaxPool2d(kernel_size=5, stride=1, padding=2)

    def forward(self, x):
        top = x
        for i in range(self.n_stages):
            top = self.maxpool(top)
            top = getattr(self, '{}_{}'.format(i + 1, 'outvar_dimred'))(top)
            x = top + x
        return x


stages_suffixes = {0: '_conv',
                   1: '_conv_relu_varout_dimred'}

class ResidualConvUnit(nn.Module):

    def __init__(self, in_planes, out_planes, n_blocks, n_stages):

        self.stride = 1
        self.n_blocks = n_blocks
        self.n_stages = n_stages

        super(ResidualConvUnit, self).__init__()
        for i in range(n_blocks):
            for j in range(n_stages):
                setattr(self, '{}{}'.format(i + 1, stages_suffixes[j]),
                        conv3x3(in_planes if (i == 0) and (j == 0) else out_planes,
                                out_planes, stride=self.stride, bias=(j == 0)))

    def forward(self, x):
        for i in range(self.n_blocks):
            residual = x
            for j in range(self.n_stages):
                x = F.relu(x)
                x = getattr(self, '{}{}'.format(i + 1, stages_suffixes[j]))(x)
            x += residual
        return x


import torch
import os, sys
from six.moves import urllib

from ..helpers import cache_location


def download_model(model_name, model_url, model_dir=None, map_location=None):
    print("Fetching pre-trained model '%s' from:\n\t%s" %
          (model_name, model_url))
    if model_dir is None:
        # Try use PyTorch's recommendations by default, otherwise fallback to
        # our package level cache
        model_dir = os.getenv('TORCH_MODEL_ZOO')
        if model_dir is None and os.getenv('TORCH_HOME') is not None:
            model_dir = os.path.join(
                os.path.expanduser(os.getenv('TORCH_HOME')), 'models')
        if model_dir is None:
            model_dir = os.path.join(cache_location(), 'pretrained')
            print("Falling back to package cache for storing "
                  "pretrained models:\n\t%s" % model_dir)
    if not os.path.exists(model_dir):
        os.makedirs(model_dir)
    filename = '{}.pth.tar'.format(model_name)
    cached_file = os.path.join(model_dir, filename)
    if not os.path.exists(cached_file):
        url = model_url
        sys.stderr.write('Downloading to {}\n'.format(cached_file))
        urllib.request.urlretrieve(url, cached_file)
    return torch.load(cached_file, map_location=map_location)


def num_classes_from_weights(weights):
    return weights['clf_conv.weight'].shape[0]



import os
import pickle
import torch
import random
import numpy as np
from tqdm import trange
from PIL import Image
from pycocotools import coco
from pycocotools import mask as cocomask
from torch.utils.data import Dataset

from .helpers import coco2voc
from ..helpers import ColourMap


class COCO(Dataset):
    '''Microsoft COCO Segmentation dataset.'''
    COLOUR_MAP = ColourMap(dataset='voc')
    LABEL_OFFSET = 0
    NUM_CLASSES = 21

    def __init__(self,
                 root_dir,
                 image_set='train',
                 transform=None,
                 target_transform=None):

        self.root_dir = root_dir
        self.image_set = image_set
        if self.image_set == 'train':
            self.img_dir = os.path.join(root_dir, 'train2017')
            self.ann_file = os.path.join(
                root_dir, 'annotations/instances_train2017.json')
            self.ids_file = os.path.join(
                root_dir, 'annotations/instances_train2017_ids.mx')
        elif self.image_set == 'val':
            self.img_dir = os.path.join(root_dir, 'val2017')
            self.ann_file = os.path.join(root_dir,
                                         'annotations/instances_val2017.json')
            self.ids_file = os.path.join(
                root_dir, 'annotations/instances_val2017_ids.mx')
        self.transform = transform
        self.target_transform = target_transform

        # gets COCO to VOC classes mapping
        self.coco2voc = coco2voc()
        self.coco = coco.COCO(self.ann_file)
        self.categories = self.coco.loadCats(self.coco.getCatIds())
        self.img_ids = self.coco.getImgIds()
        self.catIds = self.coco.getCatIds()

        # check if ids of qualified images exists, otherwise create list of qualified ids
        if os.path.exists(self.ids_file):
            with open(self.ids_file, 'rb') as f:
                self.img_ids = pickle.load(f)
        else:
            ids = list(self.coco.imgs.keys())
            self.img_ids = self.preprocess(ids, self.ids_file)

        # dataset properties
        self.num_classes = COCO.NUM_CLASSES
        self.ignore_index = 255
        self.label_offset = COCO.LABEL_OFFSET
        self.cmap = COCO.COLOUR_MAP

    def __get_year(self):
        name = self.dataset_name
        if 'coco' in name:
            name = name.replace('coco', '')
        else:
            name = name.replace('COCO', '')
        year = name
        return year

    def __len__(self):
        return len(self.img_ids)

    def __getitem__(self, idx):

        # load image data
        img_ann = self.coco.loadImgs(self.img_ids[idx])

        # get filename
        filename = img_ann[0]['file_name']

        img_name = os.path.join(self.img_dir, filename)
        image = Image.open(img_name)
        image = image.convert('RGB')
        h, w = image.height, image.width

        # load label data
        ann_ids = self.coco.getAnnIds(imgIds=self.img_ids[idx])
        anns = self.coco.loadAnns(ann_ids)
        label = np.zeros((image.height, image.width), dtype=np.uint8)
        for ann_item in anns:
            mask = self.coco.annToMask(ann_item)
            label[mask > 0] = self.coco2voc[ann_item['category_id']]

        # check segmentation is correct & convert to PIL image for transform
        if np.max(label) > 91:
            print(np.max(label))
            raise ValueError('segmentation > 91')
        if np.max(label) > 20:
            print(np.max(label))
            raise ValueError('segmentation > 20')
        label = Image.fromarray(label)

        # ensure same random transformation being applied to both data and target
        seed = np.random.randint(2147483647)
        random.seed(seed)
        if self.transform:
            image = self.transform(image)

        random.seed(seed)
        if self.target_transform:
            label = self.target_transform(label)

        # convert to label to tensor (without scaling to [0,1])
        label = np.asarray(label).astype(np.uint8)
        label = torch.from_numpy(label).type(torch.LongTensor)

        # create sample of data and label
        sample = {'name': filename, 'data': image, 'label': label}

        return sample

    def gen_seg_mask(self, target, h, w):
        mask = np.zeros((h, w), dtype=np.uint8)
        for instance in target:
            rle = cocomask.frPyObjects(instance['segmentation'], h, w)
            m = cocomask.decode(rle)
            cat = instance['category_id']
            if self.coco2voc[cat] != 0:
                c = self.coco2voc[cat]
            else:
                continue
            if len(m.shape) < 3:
                mask[:, :] += (mask == 0) * (m * c)
            else:
                mask[:, :] += (mask == 0) * ((
                    (np.sum(m, axis=2)) > 0) * c).astype(np.uint8)
        return mask

    def preprocess(self, ids, ids_file):
        tbar = trange(len(ids))
        new_ids = []
        for i in tbar:
            img_id = ids[i]
            cocotarget = self.coco.loadAnns(self.coco.getAnnIds(imgIds=img_id))
            img_metadata = self.coco.loadImgs(img_id)[0]
            mask = self.gen_seg_mask(cocotarget, img_metadata['height'],
                                     img_metadata['width'])
            if (mask > 0).sum() > 1000:
                new_ids.append(img_id)
            tbar.set_description(
                'Doing: {}/{}, got {} qualified images'.format(
                    i, len(ids), len(new_ids)))
        print('Found number of qualified images: ', len(new_ids))
        with open(ids_file, 'wb') as f:
            pickle.dump(new_ids, f)
        return new_ids



import timeit
import os
import torch
import torch.nn as nn
from torch.utils.data import DataLoader


def get_lr(optimizer):
    for param_group in optimizer.param_groups:
        return param_group['lr']


class Trainer(nn.Module):

    def __init__(self, output_directory=None):
        super().__init__()

        # Declare directory for outputs
        self.output_directory = output_directory

    def train(self, model, dataset, *, eval_interval, snapshot_interval,
              display_interval, batch_size, num_workers, freeze_batch_normal):

        # freeze batch norm during training
        if freeze_batch_normal:
            for m in model.modules():
                if isinstance(m, nn.BatchNorm2d):
                    m.eval()

        # make save directory
        if self.output_directory is not None:
            os.makedirs(self.output_directory, exist_ok=True)

        # go through each dataset
        curr_epoch = 0
        curr_iteration = 0
        for i, d in enumerate(dataset['train']):

            # create schedulers for encoder and decoder optimisers
            enc_scheduler = torch.optim.lr_scheduler.StepLR(
                model.enc_optimiser,
                step_size=dataset['stage_epochs'][i],
                gamma=dataset['stage_gammas'][i])
            dec_scheduler = torch.optim.lr_scheduler.StepLR(
                model.dec_optimiser,
                step_size=dataset['stage_epochs'][i],
                gamma=dataset['stage_gammas'][i])

            # setup dataloader
            dataloader = DataLoader(d,
                                    batch_size=batch_size,
                                    shuffle=True,
                                    num_workers=num_workers)

            # define loss criterion
            criterion = nn.NLLLoss(ignore_index=d.ignore_index)

            # training
            model.train()
            start = timeit.default_timer()
            for epoch in range(dataset['stage_epochs'][i]):

                for batch in dataloader:

                    # extract data and labels from batch
                    data = batch['data']
                    labels = batch['label']
                    if model.cuda_available:
                        data = data.cuda()
                        labels = labels.cuda()

                    # fit data to model
                    mean_loss = model.fit(data, labels, criterion)

                    # display current training loss
                    if (curr_iteration + 1) % display_interval == 0:
                        stop = timeit.default_timer()
                        print(
                            '[Epoch: {}] [Iter: {}] [Time: {:3f}] [Lr: {}] [loss: {:4f}]'
                            .format(curr_epoch, curr_iteration + 1,
                                    stop - start, get_lr(model.enc_optimiser),
                                    mean_loss.item()))
                        start = stop
                    curr_iteration += 1

                # evaluate model by computing mean IU
                if (epoch + 1) % eval_interval == 0 and dataset['val']:
                    # set model to eval first
                    model.eval()
                    mean_iu = model.validate(dataset['val'])
                    print('[Epoch: {}] [Iter: {}] [mean IU: {:4f}]'.format(
                        curr_epoch, curr_iteration + 1, mean_iu))
                    if self.output_directory is not None:
                        with open(
                                os.path.join(self.output_directory,
                                             'mean_iu.txt'), 'a') as f:
                            f.writelines([
                                'Epoch: ',
                                str(curr_epoch + 1), ' ', 'Mean IU: ',
                                str(mean_iu), '\n'
                            ])

                    # model back to train mode
                    model.train()

                # save the model
                if (epoch + 1) % snapshot_interval == 0:
                    model.save(curr_epoch + 1, self.output_directory)

                # step scheduler
                enc_scheduler.step()
                dec_scheduler.step()

                # update current epoch
                curr_epoch += 1



import os
import numpy as np
import torch
import random
from PIL import Image
from torch.utils.data.dataset import Dataset

from .helpers import read_filelist
from ..helpers import ColourMap


class VOC(Dataset):
    '''Pascal VOC Segmentation dataset.'''
    COLOUR_MAP = ColourMap(dataset='voc')
    LABEL_OFFSET = 0
    NUM_CLASSES = 21

    def __init__(self,
                 root_dir,
                 image_set='train',
                 transform=None,
                 target_transform=None):
        '''
        Args:
            csv_file (string): Path to the csv file with annotations.
            root_dir (string): Directory with all the images.
            transform (callable, optional): Optional transform to be applied
                on a sample.
        '''

        self.root_dir = root_dir
        self.image_set = image_set
        if self.image_set == 'train':
            self.file_list = read_filelist(
                os.path.join(
                    root_dir,
                    'VOCdevkit/VOC2012/ImageSets/Segmentation/train.txt'))
        elif self.image_set == 'trainval':
            self.file_list = read_filelist(
                os.path.join(
                    root_dir,
                    'VOCdevkit/VOC2012/ImageSets/Segmentation/trainval.txt'))
        elif self.image_set == 'val' or self.image_set == 'test':
            self.file_list = read_filelist(
                os.path.join(
                    root_dir,
                    'VOCdevkit/VOC2012/ImageSets/Segmentation/val.txt'))
        self.transform = transform
        self.target_transform = target_transform

        # dataset properties
        self.num_classes = VOC.NUM_CLASSES
        self.ignore_index = 255
        self.label_offset = VOC.LABEL_OFFSET
        self.cmap = VOC.COLOUR_MAP

    def __len__(self):
        return len(self.file_list)

    def __getitem__(self, idx):
        if torch.is_tensor(idx):
            idx = idx.tolist()

        # get filename
        filename = self.file_list[idx]

        # load image data and convert to ndarray
        img_name = os.path.join(self.root_dir, 'VOCdevkit/VOC2012/JPEGImages',
                                filename + '.jpg')
        image = Image.open(img_name)
        image = image.convert('RGB')

        # load label data
        label_name = os.path.join(self.root_dir,
                                  'VOCdevkit/VOC2012/SegmentationClass',
                                  filename + '.png')
        label = Image.open(label_name)

        seed = np.random.randint(2147483647)
        random.seed(seed)
        if self.transform:
            image = self.transform(image)

        random.seed(seed)
        if self.target_transform:
            label = self.target_transform(label)

        # convert to label to tensor (without scaling to [0,1])
        label = np.asarray(label).astype(np.uint8)
        label = torch.from_numpy(label).type(torch.LongTensor)

        # create sample of data and label
        sample = {'name': filename, 'data': image, 'label': label}

        return sample



import os
import torch
import torch.nn as nn
import numpy as np
from PIL import Image
import torch

from .helpers import (compute_cm, compute_iu, forward_multi_scale,
                      forward_single_scale)


class Evaluator(nn.Module):
    MIU_FILENAME = 'mean_iu.txt'

    def __init__(self,
                 multi_scale=False,
                 output_directory='.',
                 output_images=False):
        super().__init__()

        self.multi_scale = multi_scale
        self.output_directory = output_directory
        self.output_images = output_images

    # sample images using specified snapshot model
    def sample(self, model, dataset):
        # create dataloader using dataset
        dataloader = torch.utils.data.DataLoader(dataset,
                                                 batch_size=1,
                                                 shuffle=False,
                                                 num_workers=1)

        # Create the output directory for our images if required
        output_directory = (os.path.join(self.output_directory, 'images')
                            if self.output_images else None)
        if output_directory is not None:
            os.makedirs(output_directory, exist_ok=True)

        # set model to eval mode for inference
        model.eval()
        # sample images and generate prediction images
        with torch.no_grad():
            for batch in dataloader:

                # retrieve required data from batch
                name = batch['name']
                img = batch['data']
                if model.cuda_available:
                    img = img.cuda()

                # predict using single or multi-scale images
                prediction = (forward_multi_scale if self.multi_scale else
                              forward_single_scale)(model, img)

                # convert PyTorch tensor to ndarray
                prediction = prediction.cpu().detach().numpy().astype(np.uint8)

                # apply colour map
                prediction += dataset.label_offset
                prediction = dataset.cmap.colourise(prediction)

                # save prediction image if requested
                if output_directory is not None:
                    Image.fromarray(prediction).save(
                        os.path.join(output_directory, '%s.png' % name[0]))

    def compute_miu(self, model, dataset):
        # create dataloader using dataset
        dataloader = torch.utils.data.DataLoader(dataset,
                                                 batch_size=1,
                                                 shuffle=False,
                                                 num_workers=1)

        # full confusion matrix
        full_cm = torch.zeros((dataset.num_classes, dataset.num_classes),
                              dtype=torch.int64)
        if model.cuda_available:
            full_cm = full_cm.cuda()

        # set model to eval mode for inference
        model.eval()
        # sample images and generate prediction images
        with torch.no_grad():
            for batch in dataloader:

                # retrieve required data from batch
                img = batch['data']
                label = batch['label']
                if model.cuda_available:
                    img = img.cuda()
                    label = label.cuda()

                # predict using single or multi-scale images
                prediction = (forward_multi_scale if self.multi_scale else
                              forward_single_scale)(model, img)

                # compute IU
                cm = compute_cm(label, prediction, dataset.num_classes,
                                model.cuda_available)

                # add to total confusion matrix
                full_cm += cm

            # compute mean IU from confusion matrix
            full_cm = full_cm.cpu().detach().numpy().astype(np.int64)
            iu = compute_iu(full_cm)
            mean_iu = np.mean(iu)

            with open(
                    os.path.join(self.output_directory,
                                 Evaluator.MIU_FILENAME), 'a') as f:
                f.writelines(['Mean IU: ', str(mean_iu), '\n'])



import acrv_datasets
import os
import numpy as np
import PIL.Image as Image
import re
import torch
from torchvision import transforms

from .datasets.coco import COCO
from .datasets.nyu import NYU
from .datasets.sbd import SBD
from .datasets.voc import VOC
from .evaluator import Evaluator
from .models import refinenet, refinenet_lw
from .predictor import Predictor
from .trainer import Trainer


class RefineNet(object):
    # TODO should citiscapes, coco, sbd be in here???
    COLORMAP_PRESETS = {'nyu': NYU, 'voc': VOC}
    # TODO citiscapes should be in these lists!!!
    DATASETS = ['nyu', 'voc']
    DATASET_NUM_CLASSES = {'nyu': NYU.NUM_CLASSES, 'voc': VOC.NUM_CLASSES}

    MODEL_TYPES = ['full', 'lightweight']
    NUM_LAYERS = [50, 101, 152]
    OPTIMISER_TYPES = ['adam', 'sgd']
    PRETRAINED = ['imagenet', 'nyu', 'voc']

    def __init__(self,
                 *,
                 gpu_id=0,
                 model_type='full',
                 model_seed=0,
                 num_classes=21,
                 num_resnet_layers=50,
                 load_pretrained='imagenet',
                 load_snapshot=None,
                 load_snapshot_optimiser=True):
        # Apply sanitised arguments
        self.gpu_id = gpu_id
        self.model_type = _sanitise_arg(model_type, 'model_type',
                                        RefineNet.MODEL_TYPES)
        self.model_seed = model_seed
        self.num_classes = num_classes
        self.num_resnet_layers = _sanitise_arg(num_resnet_layers,
                                               'num_resnet_layers',
                                               RefineNet.NUM_LAYERS)
        self.load_pretrained = (
            None if load_pretrained is None else _sanitise_arg(
                load_pretrained, 'load_pretrained', RefineNet.PRETRAINED))
        self.load_snapshot = load_snapshot
        self.load_snapshot_optimiser = load_snapshot_optimiser

        # Try setting up GPU integration
        os.environ['CUDA_DEVICE_ORDER'] = 'PCI_BUS_ID'
        os.environ['CUDA_VISIBLE_DEVICES'] = str(self.gpu_id)
        torch.manual_seed(self.model_seed)
        if not torch.cuda.is_available():
            raise RuntimeWarning("PyTorch could not find CUDA, using CPU ...")

        # Load the model based on the specified parameters
        self.model = None
        if self.load_snapshot:
            print("\nLOADING MODEL FROM SNAPSHOT:")
            self.model = _from_snapshot(self.load_snapshot,
                                        self.load_snapshot_optimiser)
        else:
            print("\nLOADING MODEL FROM PRE-TRAINED WEIGHTS:")
            self.model = _get_model(self.model_type,
                                    self.num_resnet_layers,
                                    self.num_classes,
                                    pretrained=self.load_pretrained)
        self.model.cuda()

    def eval(self,
             dataset_name,
             *,
             dataset_dir=None,
             multi_scale=False,
             output_directory='.',
             output_images=False):
        # Perform argument validation
        dataset_name = _sanitise_arg(dataset_name, 'dataset_name',
                                     RefineNet.DATASETS)

        # Load in the dataset
        dataset = _load_dataset(dataset_name, dataset_dir, self.model_type)

        # Perform the requested evaluation
        e = Evaluator(multi_scale=multi_scale,
                      output_directory=output_directory,
                      output_images=output_images)
        e.sample(self.model, dataset['val'])
        e.compute_miu(self.model, dataset['val'])

    def predict(self,
                *,
                colour_map_preset='nyu',
                image=None,
                image_file=None,
                multi_scale=False,
                output_file=None):
        # Handle input arguments
        cmap_preset = _sanitise_arg(colour_map_preset, 'colour_map_preset',
                                    RefineNet.COLORMAP_PRESETS)
        cmap_preset = RefineNet.COLORMAP_PRESETS[cmap_preset]
        if image is None and image_file is None:
            raise ValueError("Only one of 'image' or 'image_file' can be "
                             "used in a call, not both.")
        elif image is not None and image_file is not None:
            raise ValueError("Either 'image' or 'image_file' must be provided")
        if output_file:
            os.makedirs(os.path.dirname(output_file), exist_ok=True)

        # Construct the input image
        img = (np.array(Image.open(image_file).convert('RGB'))
               if image_file else image)
        img = _get_transforms()[2](img)

        # Perform the forward pass
        out = Predictor(multi_scale=multi_scale).predict(
            img,
            self.model,
            colour_map=cmap_preset.COLOUR_MAP,
            label_offset=cmap_preset.LABEL_OFFSET)

        # Save the file if requested, & return the output
        if output_file:
            Image.fromarray(out).save(output_file)
        return out

    def train(self,
              dataset_name,
              *,
              batch_size=4,
              dataset_dir=None,
              display_interval=10,
              eval_interval=1,
              freeze_batch_normal=False,
              learning_rate=5e-4,
              num_workers=4,
              optimiser_type='sgd',
              output_directory=os.path.expanduser('~/refinenet-output'),
              snapshot_interval=5):
        # Perform argument validation / set defaults
        dataset_name = _sanitise_arg(dataset_name, 'dataset_name',
                                     RefineNet.DATASETS)
        optimiser_type = _sanitise_arg(optimiser_type, 'optimiser_type',
                                       RefineNet.OPTIMISER_TYPES)
        dataset_num_classes = RefineNet.DATASET_NUM_CLASSES.get(
            dataset_name, 0)
        if self.model.num_classes != dataset_num_classes:
            raise ValueError(
                "Can't train using a dataset with '%d' classes, when your "
                "model has been created for '%d' classes." %
                (dataset_num_classes, self.model.num_classes))

        # Load in the dataset
        dataset = _load_dataset(dataset_name, dataset_dir, self.model_type)

        # Attach new optimiser if required (and resend to GPU???)
        if (not hasattr(self.model, 'enc_optimiser') or
                not hasattr(self.model, 'dec_optimiser')):
            print("\nATTACHING NEW OPTIMISERS:")
            _attach_new_optimiser(self.model, self.model_type, optimiser_type,
                                  learning_rate)
            if self.model.cuda_available:
                self.model.cuda()

        # Start a model trainer
        print("\nPERFORMING TRAINING:")
        Trainer(output_directory).train(
            self.model,
            dataset,
            eval_interval=eval_interval,
            snapshot_interval=snapshot_interval,
            display_interval=display_interval,
            batch_size=batch_size,
            num_workers=num_workers,
            freeze_batch_normal=freeze_batch_normal)


def _attach_new_optimiser(model, model_type, optimiser_type, learning_rate):
    # Get encoder and decoder parameters from the model
    enc_params = []
    dec_params = []
    for k, v in model.named_parameters():
        if bool(re.match(".*conv1.*|.*bn1.*|.*layer.*", k)):
            enc_params.append(v)
        else:
            dec_params.append(v)

    # Attach the optimisers based on model_type & optimiser_type
    opt_fn = (torch.optim.SGD if optimiser_type == RefineNet.OPTIMISER_TYPES[1]
              else torch.optim.Adam)
    opt_params = ({
        'momentum':
            0.9,
        'weight_decay':
            5e-4 if model_type == RefineNet.MODEL_TYPES[0] else 1e-5
    } if optimiser_type == RefineNet.OPTIMISER_TYPES[1] else {})
    model.enc_optimiser = opt_fn(lr=learning_rate,
                                 params=enc_params,
                                 **opt_params)
    model.dec_optimiser = opt_fn(
        lr=(learning_rate if model_type == RefineNet.MODEL_TYPES[0] else 10 *
            learning_rate),
        params=dec_params,
        **opt_params)


def _from_snapshot(snapshot_filename, with_optim=True):
    print('Loading model from:\n\t%s' % snapshot_filename)
    full_model = torch.load(snapshot_filename)

    # Determine what class to load

    # self.load_state_dict(full_model['model'], strict=False)
    # if with_optim:
    #     self.enc_optimiser.load_state_dict(full_model['enc_optimiser'])
    #     self.dec_optimiser.load_state_dict(full_model['dec_optimiser'])
    #     # move optimiser to cuda
    #     if self.cuda_available:
    #         for state in self.optimiser.state.values():
    #             for k, v in state.items():
    #                 if isinstance(v, torch.Tensor):
    #                     state[k] = v.cuda()
    curr_iteration = full_model['global_iteration']
    return curr_iteration


def _get_model(model_type,
               num_resnet_layers,
               num_classes,
               pretrained='imagenet'):
    return {
        RefineNet.MODEL_TYPES[0]: {
            50: refinenet.refinenet50,
            101: refinenet.refinenet101,
            152: refinenet.refinenet152
        },
        RefineNet.MODEL_TYPES[1]: {
            50: refinenet_lw.refinenet_lw50,
            101: refinenet_lw.refinenet_lw101,
            152: refinenet_lw.refinenet_lw152
        }
    }[model_type][num_resnet_layers](num_classes=num_classes,
                                     pretrained=pretrained)


def _get_transforms(crop_size=224, lower_scale=1.0, upper_scale=1.0):
    # Returns 4 transforms: transform train, target transform train, transform
    # eval & target transform eval
    return (transforms.Compose([
        transforms.RandomResizedCrop(size=crop_size,
                                     scale=(lower_scale, upper_scale),
                                     interpolation=Image.BILINEAR),
        transforms.RandomHorizontalFlip(),
        transforms.ToTensor(),
        transforms.Normalize(mean=[0.485, 0.456, 0.406],
                             std=[0.229, 0.224, 0.225])
    ]),
            transforms.Compose([
                transforms.RandomResizedCrop(size=crop_size,
                                             scale=(lower_scale, upper_scale),
                                             interpolation=Image.NEAREST),
                transforms.RandomHorizontalFlip()
            ]),
            transforms.Compose([
                transforms.ToTensor(),
                transforms.Normalize(mean=[0.485, 0.456, 0.406],
                                     std=[0.229, 0.224, 0.225])
            ]), None)


def _load_dataset(dataset_name, dataset_dir, model_type, quiet=False):
    # Print some verbose information
    if not quiet:
        print("\nGETTING DATASET:")
    if dataset_dir is None:
        # TODO translate voc into all the required datasets (i.e. this
        # should handle multiple dataset_dirs)
        dataset_dir = acrv_datasets.get_datasets(dataset_name)
    if not quiet:
        print("Using 'dataset_dir': %s" % dataset_dir)

    # Get transformations
    (transform_train, target_transform_train, transform_eval,
     target_transform_eval) = _get_transforms(**({
         'crop_size': 400,
         'lower_scale': 0.7,
         'upper_scale': 1.3
     } if model_type == 'full' else {
         'crop_size': 500,
         'lower_scale': 0.5,
         'upper_scale': 2.0
     }))

    # Construct & return the dataset dictionary
    # TODO dataset dirs for VOC aren't currently handled correctly!!!
    train_args = {
        'root_dir': dataset_dir,
        'image_set': 'train',
        'transform': transform_train,
        'target_transform': target_transform_train
    }
    eval_args = {
        'root_dir': dataset_dir,
        'image_set': 'test',
        'transform': transform_eval,
        'target_transform': target_transform_eval
    }
    return ({
        'train': [NYU(**train_args)] * (2 if model_type == 'full' else 3),
        'val':
            NYU(**eval_args),
        'stage_epochs':
            ([250, 250] if model_type == 'full' else [100, 100, 100]),
        'stage_gammas':
            ([0.1, 0.1] if model_type == 'full' else [0.5, 0.5, 0.5])
    } if dataset_name == 'nyu' else {
        'train': [],
        # [COCO(**train_args),
        #  SBD(**train_args),
        #  VOC(**train_args)],
        'val': VOC(**eval_args),
        'stage_epochs': [20, 50, 200],
        'stage_gammas': ([0.1 if model_type == 'full' else 0.5] * 3)
    })


def _sanitise_arg(value, name, supported_list):
    ret = value.lower() if type(value) is str else value
    if ret not in supported_list:
        raise ValueError("Invalid '%s' provided. Supported values are one of:"
                         "\n\t%s" % (name, supported_list))
    return ret



import os
import torch
import numpy as np
from torch.utils.data import DataLoader

from .blocks import *
from .helpers import download_model, num_classes_from_weights
from ..helpers import compute_cm, compute_iu


class RefineNet(nn.Module):

    def __init__(self, block, layers, num_layers=50, num_classes=21):
        super(RefineNet, self).__init__()

        # check for cuda availability
        self.cuda_available = True if torch.cuda.is_available() else False

        # general params
        self.num_layers = num_layers
        self.num_classes = num_classes
        self.inplanes = 64

        # resnet backbone
        self.do = nn.Dropout(p=0.5)
        self.conv1 = nn.Conv2d(3,
                               64,
                               kernel_size=7,
                               stride=2,
                               padding=3,
                               bias=False)
        self.bn1 = nn.BatchNorm2d(64)
        self.relu = nn.ReLU(inplace=True)
        self.maxpool = nn.MaxPool2d(kernel_size=3, stride=2, padding=1)
        self.layer1 = self._make_layer(block, 64, layers[0])
        self.layer2 = self._make_layer(block, 128, layers[1], stride=2)
        self.layer3 = self._make_layer(block, 256, layers[2], stride=2)
        self.layer4 = self._make_layer(block, 512, layers[3], stride=2)

        # RefineNet
        # RefineNet Block 1
        self.p_ims1d2_outl1_dimred = conv3x3(2048, 512, bias=False)
        self.adapt_stage1_b = self._make_rcu(512, 512, 2, 2)
        self.mflow_conv_g1_pool = self._make_crp(512, 512, 4)
        self.mflow_conv_g1_b = self._make_rcu(512, 512, 3, 2)
        self.mflow_conv_g1_b3_joint_varout_dimred = conv3x3(512,
                                                            256,
                                                            bias=False)

        # RefineNet Block 2
        self.p_ims1d2_outl2_dimred = conv3x3(1024, 256, bias=False)
        self.adapt_stage2_b = self._make_rcu(256, 256, 2, 2)
        self.adapt_stage2_b2_joint_varout_dimred = conv3x3(256,
                                                           256,
                                                           bias=False)
        self.mflow_conv_g2_pool = self._make_crp(256, 256, 4)
        self.mflow_conv_g2_b = self._make_rcu(256, 256, 3, 2)
        self.mflow_conv_g2_b3_joint_varout_dimred = conv3x3(256,
                                                            256,
                                                            bias=False)

        # RefineNet Block 3
        self.p_ims1d2_outl3_dimred = conv3x3(512, 256, bias=False)
        self.adapt_stage3_b = self._make_rcu(256, 256, 2, 2)
        self.adapt_stage3_b2_joint_varout_dimred = conv3x3(256,
                                                           256,
                                                           bias=False)
        self.mflow_conv_g3_pool = self._make_crp(256, 256, 4)
        self.mflow_conv_g3_b = self._make_rcu(256, 256, 3, 2)
        self.mflow_conv_g3_b3_joint_varout_dimred = conv3x3(256,
                                                            256,
                                                            bias=False)

        # RefineNet Block 4
        self.p_ims1d2_outl4_dimred = conv3x3(256, 256, bias=False)
        self.adapt_stage4_b = self._make_rcu(256, 256, 2, 2)
        self.adapt_stage4_b2_joint_varout_dimred = conv3x3(256,
                                                           256,
                                                           bias=False)
        self.mflow_conv_g4_pool = self._make_crp(256, 256, 4)
        self.mflow_conv_g4_b = self._make_rcu(256, 256, 3, 2)

        self.clf_conv = nn.Conv2d(256,
                                  num_classes,
                                  kernel_size=3,
                                  stride=1,
                                  padding=1,
                                  bias=True)

    def _make_crp(self, in_planes, out_planes, stages):
        layers = [ChainedResidualPool(in_planes, out_planes, stages)]
        return nn.Sequential(*layers)

    def _make_rcu(self, in_planes, out_planes, blocks, stages):
        layers = [ResidualConvUnit(in_planes, out_planes, blocks, stages)]
        return nn.Sequential(*layers)

    def _make_layer(self, block, planes, blocks, stride=1):
        downsample = None
        if stride != 1 or self.inplanes != planes * block.expansion:
            downsample = nn.Sequential(
                nn.Conv2d(self.inplanes,
                          planes * block.expansion,
                          kernel_size=1,
                          stride=stride,
                          bias=False),
                nn.BatchNorm2d(planes * block.expansion),
            )

        layers = []
        layers.append(block(self.inplanes, planes, stride, downsample))
        self.inplanes = planes * block.expansion
        for i in range(1, blocks):
            layers.append(block(self.inplanes, planes))

        return nn.Sequential(*layers)

    def forward(self, x):

        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu(x)
        x = self.maxpool(x)

        l1 = self.layer1(x)
        l2 = self.layer2(l1)
        l3 = self.layer3(l2)
        l4 = self.layer4(l3)

        l4 = self.do(l4)
        l3 = self.do(l3)

        x4 = self.p_ims1d2_outl1_dimred(l4)
        x4 = self.adapt_stage1_b(x4)
        x4 = self.relu(x4)
        x4 = self.mflow_conv_g1_pool(x4)
        x4 = self.mflow_conv_g1_b(x4)
        x4 = self.mflow_conv_g1_b3_joint_varout_dimred(x4)
        x4 = nn.Upsample(size=l3.size()[2:],
                         mode='bilinear',
                         align_corners=True)(x4)

        x3 = self.p_ims1d2_outl2_dimred(l3)
        x3 = self.adapt_stage2_b(x3)
        x3 = self.adapt_stage2_b2_joint_varout_dimred(x3)
        x3 = x3 + x4
        x3 = F.relu(x3)
        x3 = self.mflow_conv_g2_pool(x3)
        x3 = self.mflow_conv_g2_b(x3)
        x3 = self.mflow_conv_g2_b3_joint_varout_dimred(x3)
        x3 = nn.Upsample(size=l2.size()[2:],
                         mode='bilinear',
                         align_corners=True)(x3)

        x2 = self.p_ims1d2_outl3_dimred(l2)
        x2 = self.adapt_stage3_b(x2)
        x2 = self.adapt_stage3_b2_joint_varout_dimred(x2)
        x2 = x2 + x3
        x2 = F.relu(x2)
        x2 = self.mflow_conv_g3_pool(x2)
        x2 = self.mflow_conv_g3_b(x2)
        x2 = self.mflow_conv_g3_b3_joint_varout_dimred(x2)
        x2 = nn.Upsample(size=l1.size()[2:],
                         mode='bilinear',
                         align_corners=True)(x2)

        x1 = self.p_ims1d2_outl4_dimred(l1)
        x1 = self.adapt_stage4_b(x1)
        x1 = self.adapt_stage4_b2_joint_varout_dimred(x1)
        x1 = x1 + x2
        x1 = F.relu(x1)
        x1 = self.mflow_conv_g4_pool(x1)
        x1 = self.mflow_conv_g4_b(x1)
        x1 = self.do(x1)

        out = self.clf_conv(x1)
        return out

    def fit(self, data, labels, criterion):

        # forward pass through the model
        output = self.forward(data)

        # apply dense softmax
        output = F.softmax(output, dim=1)

        # interpolate output to match label data size
        output = F.interpolate(output, (labels.shape[-2], labels.shape[-1]),
                               mode='bilinear',
                               align_corners=True)

        # apply log
        output = torch.log(output)

        # reshape label tensor to 3D (N, H, W)
        labels = labels.view(-1, labels.shape[-2], labels.shape[-1])

        # compute loss from output and labels using criterion
        loss = criterion(output, labels)

        # do a backward pass to compute gradients
        self.enc_optimiser.zero_grad()
        self.dec_optimiser.zero_grad()
        loss.backward()

        # update model
        self.enc_optimiser.step()
        self.dec_optimiser.step()

        return loss

    def validate(self, dataset):
        dataloader = DataLoader(dataset,
                                batch_size=1,
                                shuffle=False,
                                num_workers=1)

        # full confusion matrix
        full_cm = torch.zeros((dataset.num_classes, dataset.num_classes),
                              dtype=torch.int64)
        if self.cuda_available:
            full_cm = full_cm.cuda()

        with torch.no_grad():
            for batch in dataloader:
                # extract data and labels from batch
                data = batch['data']
                labels = batch['label']
                if self.cuda_available:
                    data = data.cuda()
                    labels = labels.cuda()

                # forward pass through
                logits = self.forward(data)

                # interpolate logits back to original image size
                prediction = F.softmax(logits, dim=1)
                prediction = F.interpolate(prediction,
                                           (data.shape[-2], data.shape[-1]),
                                           mode='bilinear')
                prediction = torch.argmax(prediction, dim=1)
                prediction = torch.squeeze(prediction)

                # compute confusion matrix
                cm = compute_cm(labels, prediction, dataset.num_classes,
                                self.cuda_available)
                full_cm += cm

            # compute mean IU from confusion matrix
            full_cm = full_cm.cpu().detach().numpy().astype(np.int64)
            iu = compute_iu(full_cm)
            mean_iu = np.mean(iu)

        return mean_iu

    def save(self, global_iteration, log_directory):
        os.makedirs(os.path.join(log_directory, 'snapshots'), exist_ok=True)
        model = {
            'enc_optimiser': self.enc_optimiser.state_dict(),
            'dec_optimiser': self.dec_optimiser.state_dict(),
            'global_iteration': global_iteration,
            'model_metadata': {
                'type': 'full',
                'num_layers': self.num_layers,
                'num_classes': self.num_classes
            },
            'weights': self.state_dict(),
        }

        model_path = os.path.join(
            log_directory, 'snapshots',
            'model-{:06d}.pth.tar'.format(global_iteration))
        print('Creating Snapshot: ' + model_path)
        torch.save(model, model_path)


imagenet_urls = {
    'resnet50': 'https://download.pytorch.org/models/resnet50-19c8e357.pth',
    'resnet101': 'https://download.pytorch.org/models/resnet101-5d3b4d8f.pth',
    'resnet152': 'https://download.pytorch.org/models/resnet152-b121ed2d.pth'
}

pretrained_urls = {
    'refinenet50_nyu':
        'https://cloudstor.aarnet.edu.au/plus/s/thgUqW2uRZBeEXG/download',
    'refinenet101_nyu':
        'https://cloudstor.aarnet.edu.au/plus/s/9wPfvg6eKqTztNy/download',
    'refinenet152_nyu':
        'https://cloudstor.aarnet.edu.au/plus/s/QFcL2qgNsnpSOlr/download',
    'refinenet50_voc':
        'https://cloudstor.aarnet.edu.au/plus/s/YH73jIiBa8iKllX/download',
    'refinenet101_voc':
        'https://cloudstor.aarnet.edu.au/plus/s/rzgoGwTb6CK7nbT/download',
    'refinenet152_voc':
        'https://cloudstor.aarnet.edu.au/plus/s/fSsPeSsNyyHYTwC/download',
}


# creates a ResNet-50 RefineNet (supports loading of pretrained ImageNet model)
def refinenet50(num_classes, pretrained='imagenet', **kwargs):
    return _refinenet(num_classes,
                      50, [3, 4, 6, 3],
                      pretrained=pretrained,
                      **kwargs)


# creates a ResNet-101 RefineNet (supports loading of pretrained ImageNet model)
def refinenet101(num_classes, pretrained='imagenet', **kwargs):
    return _refinenet(num_classes,
                      101, [3, 4, 23, 3],
                      pretrained=pretrained,
                      **kwargs)


# creates a ResNet-152 RefineNet (supports loading of pretrained ImageNet model)
def refinenet152(num_classes, pretrained='imagenet', **kwargs):
    return _refinenet(num_classes,
                      152, [3, 8, 36, 3],
                      pretrained=pretrained,
                      **kwargs)


def _refinenet(num_classes, num_resnet_layers, layers, pretrained, **kwargs):
    # Load in an appropriate set of pre-trained weights, falling back to resnet
    # if required
    map_location = None
    if not torch.cuda.is_available():
        map_location = torch.device('cpu')

    imagenet = pretrained not in ['nyu', 'voc']
    key = ('resnet%d' % num_resnet_layers if imagenet else 'refinenet%d_%s' %
           (num_resnet_layers, pretrained))
    url = (imagenet_urls if imagenet else pretrained_urls)[key]
    model = download_model(key, url, map_location=map_location)
    weights = model if imagenet else model['model']

    # Create & return a new RefineNet instance
    if not imagenet and num_classes_from_weights(weights) != num_classes:
        raise ValueError(
            "Cannot use pre-trained weights for network with '%d' classes "
            "when requesting a new network with '%d' classes" %
            (num_classes_from_weights(weights), num_classes))
    model = RefineNet(Bottleneck,
                      layers,
                      num_resnet_layers,
                      num_classes=num_classes,
                      **kwargs)
    model.load_state_dict(weights)
    model.name = key
    return model



from setuptools import Extension, find_packages, setup
from setuptools.command.install import install
import subprocess

import time

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(name='refinenet',
      version='0.9.0',
      author='Ben Talbot',
      author_email='b.talbot@qut.edu.au',
      description='RefineNet semantic segmentation',
      long_description=long_description,
      long_description_content_type='text/markdown',
      packages=find_packages(),
      install_requires=[
          'acrv_datasets', 'numpy', 'pillow', 'pycocotools', 'scipy', 'six',
          'torch', 'torchvision'
      ],
      classifiers=(
          "Development Status :: 4 - Beta",
          "Programming Language :: Python :: 3",
          "License :: OSI Approved :: BSD License",
          "Operating System :: OS Independent",
      ))



import os
import torch
import torch.nn as nn
import numpy as np

from .helpers import forward_multi_scale, forward_single_scale


class Predictor(nn.Module):

    def __init__(self, multi_scale=False):
        super().__init__()

        self.multi_scale = multi_scale

    # sample images using specified snapshot model
    def predict(self, image, model, colour_map, label_offset):
        with torch.no_grad():
            # move image to GPU
            img = image.cuda() if model.cuda_available else image

            # turn single image into NCHW format
            if len(img.shape) < 4:
                img = torch.unsqueeze(img, 0)

            # predict using single or multi-scale images
            prediction = (forward_multi_scale if self.multi_scale else
                          forward_single_scale)(model, img)

            # convert PyTorch tensor to ndarray
            prediction = prediction.cpu().detach().numpy().astype(np.uint8)

            # Return prediction with colour map applied
            return colour_map.colourise(prediction + label_offset)
